/* ###################################################################
   ## E x e r c i s e  T h r e e: model answer                      ##
   ################################################################### */
#include "bsp.h"
#include "bsp_level1.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

void spmd_start();
void total_exchange_array(int*,int,int*,int*);
int *total_exchange_int(int);

char usage_str[] = "main: -v -p <number> -s <no elements>";
int nprocs =2;
int size   =5;
int verbose=0;

/* ####################################################################
   ## The procedure "total_exchange_array" performs an all-to-all    ##
   ## total exchange of an array "src" of size "src_size" held       ##
   ## locally on each process. The value of "src_array"  may be      ##
   ## different on each process. The effect of the procedure is each ##
   ## process receives a copy of each other processes version of src,##
   ## which  are concatendated together, in process id order, into   ##
   ## the array "dst". The size of "dst", is "dst_array". However,   ##
   ## exit from the procedure, the reference parameter dst_size is   ##
   ## changed to record the actual size of the concatenated arrays.  ##
   ## An error is raised if this is larger than the value of         ##
   ## "dst_array" passed as an argument to the procedure.            ##
   #################################################################### */

void total_exchange_array(int *src, int src_size, int *dst, int *dst_size) {
}

/* ###################################################################
   ## "spmd_start" is a test harness which starts nprocs processes, ##
   ## and then does a total exchange of an array of "size", where   ##
   ## each element of the array contains the process ID.            ##
   ################################################################### */

void spmd_start() {
  int i, j, *src, *dst, max_size;
  double delta;

  bsp_begin(nprocs);
  /* The global variables size and verbose are not inherented by the
     spawned processes. We therefore broadcast them from process zero
     using the collective communication broadcast operation bsp_bcast */

    bsp_push_reg(&size,sizeof(int));
    bsp_push_reg(&verbose,sizeof(int));
    bsp_sync();
    
    bsp_bcast(0,&size,   &size,   sizeof(int));
    bsp_bcast(0,&verbose,&verbose,sizeof(int));

    /* Make sure arrays are different sizes */
    size += bsp_pid();
    if (verbose) 
      printf("Allocating %d elements on process %d\n",size,bsp_pid());

    /* Approx size of destination array */
    max_size = (bsp_nprocs() + size) * bsp_nprocs();

    /* Allocate source and destination arrays to be total exchanged */
    src  = calloc(size,sizeof(int));
    dst  = calloc(max_size,sizeof(int));
    if (src==NULL || dst==NULL)
      bsp_abort("Unable to allocate array containing %d values",size);


    /* Initialise the src array in such a way that the total exchanged array
       contains bsp_pip() at index i */
    for(i=0;i<size;i++) src[i] = bsp_pid();

    delta = bsp_dtime();
    total_exchange_array(src,size,dst,&max_size);
    
    if (bsp_pid()==0)
      printf("Practice = %f secs\nTheory   = %f secs\n\n",
             bsp_dtime(), (bsp_l() + (max_size*bsp_g())) / bsp_s());
	     
    if (verbose) {
      for(i=0;i<bsp_nprocs();i++) {
        if (bsp_pid()==i) {
          printf("On process %d dst=",i);
          for(j=0;j<max_size;j++)
            printf("%d,",dst[j]);
          printf("\n");
        }
        bsp_sync();
      }
    }  

  bsp_end();
}


/* ###################################################################
   ## The main procedure reads in the number of processes to spawn  ##
   ## off the command line. It then enters SPMD mode inside the     ##
   ## procedure spmd_start                                          ##
   ################################################################### */
    
void main(int argc,char **argv) {
  int i,temp, command_line_opts=1;
  double trash;

  /* The following procedure informs the BSPlib system that 
     the program should enter a SPMD parallel mode of execution in
     the routine spmd_start */

  bsp_init(spmd_start,argc,argv);
  i=1;
  while (command_line_opts) {
    if (i >= argc) {
      command_line_opts = 0;
    } else if ( (strcmp(argv[i],"-p")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      nprocs = temp;
      i += 2;
    }else if ( (strcmp(argv[i],"-s")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      size = temp;
      i += 2;
    } else if (strcmp(argv[i],"-v")==0) {
      verbose = 1;
      i++;
    } else {
      bsp_abort("{%s}: unknown option \"%s\"\n\t%s\n",
                argv[0],argv[i],usage_str);
    }
  }
  printf("Total exchanging %d elements between %d processes\n",size,nprocs);
  spmd_start();
  exit(0);
}

