/* ###################################################################
   ## E x e r c i s e  O n e: model answer                          ##
   ################################################################### */

#include "bsp.h"
#include <stdio.h>
#include <stdlib.h>

/* Forward declarations */
void spmd_start();
int *total_exchange_int(int);

char usage_str[] = "main: -p <number>";
int nprocs=2;

/* ###################################################################
   ## The function "total_exchange_int" performs an all-to-all      ##
   ## total exchange of a single integer held locally on each       ##
   ## process. The result of the function is that each process will ##
   ## return an array containing P integers, such that the i^th     ##
   ## element of the array contains the integer x that was passed   ##
   ## as an argument to the function on process i.                  ##
   ################################################################### */

int *total_exchange_int(int x) {
  int i,*result;

  result = calloc(bsp_nprocs(),sizeof(int));
  if (result==NULL) 
    bsp_abort("{total_exchange_int} unable to allocate %d bytes",
              bsp_nprocs()*sizeof(int));

  bsp_push_reg(result,bsp_nprocs()*sizeof(int));
  bsp_sync();

  for(i=0;i<bsp_nprocs();i++)
    bsp_put(i,&x,result,bsp_pid()*sizeof(int),sizeof(int));

  bsp_sync();
  bsp_pop_reg(result);
  return result;
}


/* ###################################################################
   ## "spmd_start" is a test harness which starts nprocs processes, ##
   ## and then does a total exchange of the ID's of each prcocess.  ##
   ################################################################### */

void spmd_start() {
  int i, j, val, *vals;

  bsp_begin(nprocs);
    val  = bsp_pid();
    bsp_dtime(); /* Start timer */

    vals  = total_exchange_int(val);

    /* Cost of the toral exchange is l + pg */
    if (bsp_pid()==0)
      printf("Practice = %f secs\nTheory   = %f secs\n\n",
             bsp_dtime(), (bsp_l() + (bsp_nprocs()*bsp_g())) / bsp_s());

    /* Make sure that the processes print out their results without 
       interleaving the IO requests */

    for(i=0;i<bsp_nprocs();i++) {
      if (bsp_pid()==i) {
        printf("On process %d vals=",i);
        for(j=0;j<bsp_nprocs();j++)
          printf("%d,",vals[j]);
        printf("\n");
      }
      bsp_sync();
      fflush(stdout);
    }
  bsp_end();
}

    
/* ###################################################################
   ## The main procedure reads in the number of processes to spawn  ##
   ## off the command line. It then enters SPMD mode inside the     ##
   ## procedure spmd_start                                          ##
   ################################################################### */

void main(int argc,char **argv) {
  int i,temp, command_line_opts=1;

  /* The following procedure informs the BSPlib system that 
     the program should enter a SPMD parallel mode of execution in
     the routine spmd_start */

  bsp_init(spmd_start,argc,argv);
  i=1;
  while (command_line_opts) {
    if (i >= argc) {
      command_line_opts = 0;
    } else if ( (strcmp(argv[i],"-p")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      nprocs = temp;
      i += 2;
    } else {
      bsp_abort("{%s}: unknown option \"%s\"\n\t%s\n",
                argv[0],argv[i],usage_str);
    }
  }
  printf("Test of bsp_begin: only one process should print this\n");
    spmd_start();
  printf("Just about to finish, only process zero should "
         "get here (bsp_pid()=%d)\n",bsp_pid());
  exit(0);
}

