C   ###################################################################
C   ## E x e r c i s e  T h r e e: model answer                      ##
C   ################################################################### 


C   ####################################################################
C   ## The procedure "total_exchange_array" performs an all-to-all    ##
C   ## total exchange of an array "src" of size "src_size" held       ##
C   ## locally on each process. The value of "src_array"  may be      ##
C   ## different on each process. The effect of the procedure is each ##
C   ## process receives a copy of each other processes version of src,##
C   ## which  are concatendated together, in process id order, into   ##
C   ## the array "dst". The size of "dst", is "dst_array". However,   ##
C   ## exit from the procedure, the reference parameter dst_size is   ##
C   ## changed to record the actual size of the concatenated arrays.  ##
C   ## An error is raised if this is larger than the value of         ##
C   ## "dst_array" passed as an argument to the procedure.            ##
C   #################################################################### 

        SUBROUTINE total_exchange_array(src,src_size,dst,dst_size)
	  INCLUDE 'fbsp.h'
          INTEGER src(*),dst(*),src_size,dst_size
          INTEGER MAXP,SIZEINT
          PARAMETER (MAXP=1024,SIZEINT=4)
          INTEGER i,p,all_sizes(MAXP), total_size, offset

          p = bspnprocs()
        
          CALL bsppushreg(dst,dst_size*SIZEINT)
          CALL bspsync()

C         Perform a basic total exchange to calculate the size of 
C         src_size on each process 
          CALL total_exchange_int(src_size,all_sizes)
          
C         From the gathered information, we then calculate the total 
C         size of the concatenated array,

          total_size = 0
          DO i=1,p
            total_size = total_size + all_sizes(i)
          END DO
C         , _and_ each process can determine the offset in the remote copy
C         of dst, where the src needs to be transfered into.
          offset = 0
          DO i=1,bsppid()
            offset = offset + all_sizes(i)
          END DO
          
          DO i=0,bspnprocs()-1
            CALL bspput(i,src,dst,offset*SIZEINT,src_size*SIZEINT)
          END DO
          CALL bspsync()
          CALL bsppopreg(dst)
          dst_size = total_size
        END

        SUBROUTINE total_exchange_int(x,xs)
	  INCLUDE 'fbsp.h'
          INTEGER x,xs(*)
          INTEGER SIZEINT
          PARAMETER (SIZEINT=4)
          INTEGER i
        
          CALL bsppushreg(xs,bspnprocs()*SIZEINT)
          CALL bspsync()

          DO i=0,bspnprocs()-1
            CALL bspput(i,x,xs,bsppid()*SIZEINT,SIZEINT)
          END DO
          CALL bspsync()
          CALL bsppopreg(xs)
        END

C   ###################################################################
C   ## The main procedure starts as many processes as possible, and  ##
C   ## then performs a total exchange of the ID's of each prcocess.  ##
C   ###################################################################

	PROGRAM MAIN
	INCLUDE 'fbsp.h'
        INTEGER MAXP,STARTSIZE,SIZE,MAX_SIZE
	PARAMETER (MAXP=1024,STARTSIZE=4)
        INTEGER src(STARTSIZE+MAXP), dst((STARTSIZE+MAXP)*MAXP)
        DOUBLE PRECISION delta
        
        CALL bspbegin(bspnprocs())

          SIZE     = STARTSIZE + bsppid()
          MAX_SIZE = (bspnprocs()+SIZE)*bspnprocs()
          WRITE (*,*) 'Allocating ',SIZE,' elements on process ',bsppid()
          DO i=1,SIZE
            src(i) = bsppid()
          END DO
          delta = bspdtime()
          CALL total_exchange_array(src,SIZE,dst,MAX_SIZE)

          DO i = 0,bspnprocs()
            IF (bsppid() .EQ. i) THEN
              WRITE (*,*) 'On process ',i , ' dst = ',
     +                    (dst(j),j=1,MAX_SIZE)
            END IF
            CALL bspsync()
          END DO

	CALL bspend()
	END
